; ==================================================================================================
; Title:   XCustomTreeView.inc
; Author:  G. Friedrich
; Version: 1.0.0
; Purpose: Customized XTreeView and XTreeViewNode.
; Notes:   Version 1.0.0, July 2007
;            - First release.
; ==================================================================================================


XTVNS_CHECKED       equ   BIT16

XTV_ICON_SIZE       equ   16
XTV_CHECK_SIZE      equ   13

XTVN_CUSTOM_DISP_INFO struct            ;Extend the XTVN_DISP_INFO structure
              XTVN_DISP_INFO    {}      ;Anonymous definition
  CheckRect   RECT              {}      ;New RECT structure for the check box
XTVN_CUSTOM_DISP_INFO ends


Object XCustomTreeView, 0, XTreeView
  RedefineMethod    Drop,               Pointer, Pointer, dword
  StaticMethod      OnCommand,          dword, dword
  RedefineMethod    OnCreate,           dword, dword                  ;Create icon
  RedefineMethod    OnDestroy,          dword, dword                  ;Destroy icon
  RedefineMethod    Startup                                           ;Register new wnd class
  RedefineMethod    WndProc,            dword, dword, dword           ;Create new wnd class

  RedefineVariable  dDispInfoSize,      sizeof XTVN_CUSTOM_DISP_INFO  ;Extended structure
  DefineVariable    hIcon,              Handle,      0

  Event   OnCommand,  WM_COMMAND
ObjectEnd


Object XCustomNode, 0, XTreeViewNode
  RedefineMethod    CanDragMe,          dword
  RedefineMethod    ContextPopup,       dword, dword
  RedefineMethod    MouseHit,           dword, Pointer, Pointer, dword;Check for check box hit
  RedefineMethod    RenderContent,      Pointer                       ;Render check box

ObjectEnd


; ==================================================================================================

if IMPLEMENT

; ==================================================================================================
;     XCustomNode
; ==================================================================================================

CStr szXCTV_Class, "XCustomTreeView"

;##\

; 
; Method:    XCustomNode.CanDragMe
; Purpose:   Indicates if a drag operation can be started.
; Arguments: Arg1: Key flags (MK_CONTROL, MK_LBUTTON, MK_MBUTTON, MK_RBUTTON, MK_SHIFT).
; Return:    TRUE / FALSE.

Method XCustomNode.CanDragMe, NOFRAME, dKeyFlags:dword
    mov eax, TRUE
MethodEnd 8, eax

; 
; Method:    XCustomNode.ContextPopup
; Purpose:   Displays a context popup menu.
; Arguments: Arg1: x coord in screen coords.
;            Arg2: y coord in screen coords.
; Return:    Nothing.

Method XCustomNode.ContextPopup, uses edi esi, dXPos:dword, dYPos:dword
    local hMenu:Handle

    SetObject esi
;    DbgHex esi
;    DbgHex [esi].pParent
;    DbgHex [esi].pFirstChild
;    DbgHex [esi].pLastChild
;    DbgHex [esi].pPrevSibling
;    DbgHex [esi].pNextSibling
;    DbgDec [esi].dBranchHeight
;    DbgDec [esi].dBranchWidth
;    mov edi, [esi].pOwner
;    DbgDec [edi].XTreeView.TotalSize.x
;    DbgDec [edi].XTreeView.TotalSize.y
;    DbgLine
    mov hMenu, $invoke(LoadMenu, hInstance, "MENU_CUSTOM_NODE")
    invoke GetSubMenu, hMenu, 0
    mov ecx, [esi].pOwner
    invoke TrackPopupMenu, eax, TPM_LEFTALIGN, dXPos, dYPos, 0, [ecx].XTreeView.hWnd, NULL
    invoke DestroyMenu, hMenu
MethodEnd, NONE

; 
; Method:    XCustomNode.MouseHit
; Purpose:   Checks if the mouse has clicked a hotspot and performs an action.
; Arguments: Arg1: button that triggered this action.
;            Arg2: -> clicked point.
;            Arg3: -> XTVN_DISP_INFO.
;            Arg4: Key flags (MK_CONTROL, MK_LBUTTON, MK_MBUTTON, MK_RBUTTON, MK_SHIFT).
; Return:    eax = TRUE if and action was triggered, otherwise FALSE.

Method XCustomNode.MouseHit, uses ebx esi, dButton:dword, pPoint:Pointer, pDispInfo:Pointer, \
                                           dKeyFlags:dword
    .if pDispInfo != NULL
      SetObject esi
      ACall esi.MouseHit, dButton, pPoint, pDispInfo, dKeyFlags
      .if eax == FALSE
        .if [esi].pFirstChild == NULL     ;Check if this is a terminal node
          mov ebx, [esi].pOwner           ;Any other node has no checkbox

          ;Check for text hit
          mov edx, pPoint
          mov eax, [edx].POINT.x
          mov ecx, [edx].POINT.y
          mov edx, pDispInfo
          assume edx:ptr XTVN_CUSTOM_DISP_INFO, ebx:ptr XCustomTreeView
          .if XTV_IsPointInRect(eax, ecx, [edx].CheckRect)
            ;DbgText "Hit on checkbox"
            mov eax, DFCS_BUTTONCHECK
            .ifBitSet [esi].dState, XTVNS_CHECKED
              BitClr [esi].dState, XTVNS_CHECKED
            .else
              BitSet [esi].dState, XTVNS_CHECKED
              or eax, DFCS_CHECKED
            .endif

            ;Render the checkbox
            invoke DrawFrameControl, [ebx].hMemDC, addr [edx].CheckRect, DFC_BUTTON, eax
            mov edx, pDispInfo
            invoke InvalidateRect, [ebx].hWnd, addr [edx].CheckRect, FALSE
            invoke UpdateWindow, [ebx].hWnd
            mov eax, TRUE
          .else
            xor eax, eax
          .endif
          assume edx:NOTHING, ebx:NOTHING
        .else
          xor eax, eax
        .endif
      .else
        xor eax, eax
      .endif
    .endif
MethodEnd, eax

; 
; Method:    XCustomNode.RenderContent
; Purpose:   Draws the content (icon + text) of the node.
; Arguments: Arg1: -> XTVN_CUSTOM_DISP_INFO
; Return:    Nothing.
; Note:      This method must update the coords of all Rect structures in the Content and must
;            invalidate the modified region.

Method XCustomNode.RenderContent, uses ebx edi esi, pDispInfo:Pointer
    local dPrevColor:dword, dPrevBkColor:dword, dPrevBkMode:dword, InvRect:RECT, hFontPrev:Handle
    local bBuffer[20]:byte

    ;DbgText "XCustomNode.RenderContent"
    mov edi, pDispInfo
    .if edi != NULL
      SetObject esi
      mov ebx, [esi].pOwner
      assume edi:ptr XTVN_CUSTOM_DISP_INFO, ebx:ptr XCustomTreeView
      m2m [edi].TextRect.left, [edi].ContentRect.left, eax
      m2m [edi].TextRect.right, [edi].ContentRect.right, eax

      ;Get the drawing rect
      m2m InvRect.left, [edi].ContentRect.left, eax
      m2m InvRect.top, [edi].sdNodeTop, eax
      m2m InvRect.right, [ebx].ClientSize.x, eax
      m2m InvRect.bottom, [edi].sdNodeBot, eax

      ;Clear the background
      OCall ebx::XTreeView.RenderBackground, addr InvRect

      ;Put an icon on nodes that have childs, else a checkbox
      .if [esi].pFirstChild == NULL
        ;Calc CheckRect
        mov eax, [edi].ContentRect.bottom
        sub eax, [edi].ContentRect.top
        sub eax, XTV_CHECK_SIZE
        shr eax, 1
        mov ecx, [edi].ContentRect.bottom
        sub ecx, eax
        mov [edi].CheckRect.bottom, ecx                 ;CheckRect.bottom
        sub ecx, XTV_CHECK_SIZE
        mov [edi].CheckRect.top, ecx                    ;CheckRect.top

        mov edx, [edi].ContentRect.left
        add edx, 3
        mov [edi].CheckRect.left, edx                   ;CheckRect.left
        add edx, XTV_CHECK_SIZE
        mov [edi].CheckRect.right, edx                  ;CheckRect.right
        add edx, 3
        mov [edi].TextRect.left, edx                    ;TextRect.left

        ;Set the control draw flags
        mov eax, DFCS_BUTTONCHECK
        .ifBitSet [esi].dState, XTVNS_CHECKED
          or eax, DFCS_CHECKED
        .endif

        ;Render the checkbox
        invoke DrawFrameControl, [ebx].hMemDC, addr [edi].CheckRect, DFC_BUTTON, eax

      .else
        ;Calc icon position
        mov eax, [edi].ContentRect.bottom
        sub eax, [edi].ContentRect.top
        sub eax, XTV_ICON_SIZE
        shr eax, 1
        mov ecx, [edi].ContentRect.bottom
        sub ecx, eax
        sub ecx, XTV_ICON_SIZE
        mov [edi].CheckRect.top, ecx                    ;ecx = Icon top

        mov edx, [edi].ContentRect.left
        add edx, 3                                      ;edx = Icon left
        mov eax, edx
        add eax, 3
        add eax, XTV_ICON_SIZE
        mov [edi].TextRect.left, eax                    ;TextRect.left

        ;Render the icon
        invoke DrawIconEx, [ebx].hMemDC, edx, ecx, [ebx].hIcon, \
                           XTV_ICON_SIZE, XTV_ICON_SIZE, NULL, NULL, DI_NORMAL

      .endif

      ;Create a text
      invoke StrECopy, addr bBuffer, $OfsCStr(" Node @ ")
      push eax
      invoke dword2hex, eax, esi
      pop eax
      add eax, 8
      @invoke StrCopy, eax, "h "

      ;Set text and background colors
      ;  Check if the whole XTV or the single node is disabled
      .if ([ebx].dState & XTVS_DISABLED) || ([esi].dState & XTVNS_DISABLED)
        ;Disabled
        invoke GetSysColor, COLOR_BTNHILIGHT
        mov dPrevColor, $invoke(SetTextColor, [ebx].hMemDC, eax)
        mov dPrevBkMode, $invoke(SetBkMode, [ebx].hMemDC, TRANSPARENT)
      .else
        ;Enabled
        .if esi == [ebx].pSelectedNode
          .ifBitSet [ebx].dState, XTVS_FOCUSED
            .ifBitSet [esi].dState, XTVNS_HOTTRACK      ;Check if the node is beeing hottracked
              mov dPrevColor, $invoke(SetTextColor, [ebx].hMemDC, $RGB(255,255,0))
            .else
              invoke GetSysColor, COLOR_HIGHLIGHTTEXT
              mov dPrevColor, $invoke(SetTextColor, [ebx].hMemDC, eax)
            .endif
            invoke GetSysColor, COLOR_HIGHLIGHT
            mov dPrevBkColor, $invoke(SetBkColor, [ebx].hMemDC, eax)
            mov dPrevBkMode, $invoke(SetBkMode, [ebx].hMemDC, OPAQUE)
          .else
            invoke GetSysColor, COLOR_WINDOWTEXT
            mov dPrevColor, $invoke(SetTextColor, [ebx].hMemDC, eax)
            invoke GetSysColor, COLOR_BTNFACE
            mov dPrevBkColor, $invoke(SetBkColor, [ebx].hMemDC, eax)
            mov dPrevBkMode, $invoke(SetBkMode, [ebx].hMemDC, OPAQUE)
          .endif
        .elseIfBitSet [esi].dState, XTVNS_HOTTRACK      ;Check if the node is beeing hottracked
          mov dPrevColor, $invoke(SetTextColor, [ebx].hMemDC, $RGB(0,0,255))
          mov dPrevBkMode, $invoke(SetBkMode, [ebx].hMemDC, TRANSPARENT)
        .else
          invoke GetSysColor, COLOR_WINDOWTEXT
          mov dPrevColor, $invoke(SetTextColor, [ebx].hMemDC, eax)
          mov dPrevBkMode, $invoke(SetBkMode, [ebx].hMemDC, TRANSPARENT)
        .endif
      .endif

      ;Set the font
      mov hFontPrev, $invoke(SelectObject, [ebx].hMemDC, [ebx].hFontText)

      ;Render the created text
      OCall esi.RenderText, edi, addr bBuffer

      ;Restore previous settings
      invoke SelectObject, [ebx].hMemDC, hFontPrev
      invoke SetBkMode, [ebx].hMemDC, dPrevBkMode
      invoke SetBkColor, [ebx].hMemDC, dPrevBkColor
      invoke SetTextColor, [ebx].hMemDC, dPrevColor

      ;Invalidate the new drawn rect
      invoke InvalidateRect, [ebx].hWnd, addr InvRect, FALSE
      assume edi:NOTHING, ebx:NOTHING
    .endif
MethodEnd, NONE


;##/

; ==================================================================================================
;     XCustomTreeView
; ==================================================================================================

;##\

; 
; Method:    XCustomTreeView.Drop
; Purpose:   Drops a node.
; Arguments: Arg1: -> Node to drop.
;            Arg2: -> Reference node.
;            Arg3: how to drop.
; Return:    Nothing.

Method XCustomTreeView.Drop, uses esi, pNode:Pointer, pRefNode:Pointer, dHowToDrop:dword
    SetObject esi
    @invoke PlaySound, "C:\Masm32\ObjAsm32\Resources\WAVs\Bleep.wav", NULL, SND_FILENAME or SND_ASYNC
    ACall esi.Drop, pNode, pRefNode, dHowToDrop
MethodEnd

; 
; Method:    XCustomTreeView.OnCommand
; Purpose:   Event handler for WM_COMMAND message.
; Arguments: Arg1: First message parameter.
;            Arg2: Second message parameter.
; Return:    WM_COMMAND return value.

Method XCustomTreeView.OnCommand,, wParam:dword, lParam:dword
    SetObject ecx
    mov eax, [ecx].pOwner
    invoke SendMessage, [eax].SdiApp.hWnd, WM_COMMAND, wParam, lParam
MethodEnd, eax

; 
; Method:    XCustomTreeView.OnCreate
; Purpose:   Event handler for WM_CREATE message.
; Arguments: Arg1: First message parameter.
;            Arg2: Second message parameter.
; Return:    eax = Zero to continue creation of the window, else -1.

Method XCustomTreeView.OnCreate, uses esi, wParam:dword, lParam:dword
    ;DbgText "XCustomTreeView.OnCreate"
    SetObject esi
    mov [esi].hIcon, $invoke(LoadIcon, hInstance, "ICON_APP_32")
;    mov [esi].hIcon, $invoke(LoadImage, hInstance, "ICON_APP_32", IMAGE_ICON, 0, 0, 80h)
    ACall esi.OnCreate, wParam, lParam
MethodEnd, eax

; 
; Method:    XCustomTreeView.OnDestroy
; Purpose:   Event handler for WM_DESTROY message.
; Arguments: Arg1: First message parameter.
;            Arg2: Second message parameter.
; Return:    Zero.

Method XCustomTreeView.OnDestroy, uses esi, wParam:dword, lParam:dword
    ;DbgText "XCustomTreeView.OnDestroy"
    SetObject esi
    invoke DestroyIcon, [esi].hIcon
    ACall esi.OnDestroy, wParam, lParam
MethodEnd, eax

; 
; Method:    XCustomTreeView.WindowProc
; Purpose:   Processing of window messages.
; Arguments: Arg1: Message identifier.
;            Arg2: First message parameter.
;            Arg3: Second message parameter.
;            Window Handle is passed in Self.
; Return:    This value is the result of the message processing and depends on the message.

Method XCustomTreeView.WndProc, uses esi, uMsg:dword, wParam:dword, lParam:dword
    ;DbgMessage uMsg
    .if uMsg == WM_NCCREATE
      mov eax, lParam
      .if [eax].CREATESTRUCT.lpCreateParams == NULL
        ;if lpCreateParams is 0, then it is a resource call => create new object instance

        mov esi, $New(XCustomTreeView, InitRes, Self)
      .else
        ;Init call
        mov esi, [eax].CREATESTRUCT.lpCreateParams
        m2m [esi].XCustomTreeView.hWnd, Self
      .endif
      invoke SetWindowLong, Self, GWL_USERDATA, esi
      OCall esi::XCustomTreeView.Dispatch, Self, uMsg, wParam, lParam
    .else
      invoke GetWindowLong, Self, GWL_USERDATA              ;if GWL_USERDATA wasn't set,
      .if eax == NULL                                       ;GetWindowLong returns NULL
        invoke DefWindowProc, Self, uMsg, wParam, lParam
      .else
        OCall eax::XCustomTreeView.Dispatch, Self, uMsg, wParam, lParam
      .endif
    .endif
MethodEnd, eax

; 
; Method:    XCustomTreeView.Startup
; Purpose:   Registers the XCustomTreeView object with the OS.
; Arguments: None.
; Return:    Nothing.

Method XCustomTreeView.Startup
    local WC:WNDCLASSEX

    mov WC.cbSize, sizeof WNDCLASSEX
    mov WC.style, CS_HREDRAW or CS_VREDRAW or CS_PARENTDC or CS_BYTEALIGNWINDOW or \
                  CS_BYTEALIGNCLIENT; or CS_DBLCLKS
    m2m WC.lpfnWndProc, $MethodAddr(XCustomTreeView.WndProc)
    mov WC.cbClsExtra, 0
    mov WC.cbWndExtra, 0
    m2m WC.hInstance, hInstance
    mov WC.hbrBackground, 0
    mov WC.lpszMenuName, NULL
    mov WC.lpszClassName, offset szXCTV_Class
    mov WC.hIcon, 0
    mov WC.hCursor, $invoke(LoadCursor, hInstance, "CURSOR_ARROW")
    mov WC.hIconSm, 0

    invoke RegisterClassEx, addr WC
MethodEnd, NONE


;##/

endif